#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#define HAND_SIZE	7
#define BOARD_SIZE	15
#define BONUS_SCORE	50
#define BAG_SIZE	100
#define BAG_VALUE	187

#define MAX_WORD	32
#define ENTRY_COUNT	100
#define POOL_SIZE	5
#define LEXICON_SIZE	180000
#define WORD_TREE_SIZE	400000
#define ALPHABET	26
#define WORD_SIZE	(BOARD_SIZE + 1)
#define DOUBLE_SIZE	(2 * BOARD_SIZE)

#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

const int allow_blanks = 2;

static int lexicon_size = 0;
static char lexicon[LEXICON_SIZE][WORD_SIZE];

static char bag[BAG_SIZE + 1];

typedef struct { int x, y; } Pos;

#define POS(X, Y) ((Pos) { .x = X, .y = Y })

static Pos add(Pos a, Pos b) __attribute__ ((pure));
static Pos sub(Pos a, Pos b) __attribute__ ((pure));

static Pos add(Pos a, Pos b) {
    return POS(a.x + b.x, a.y + b.y);
}

static Pos sub(Pos a, Pos b) {
    return POS(a.x - b.x, a.y - b.y);
}

static Pos inv(Pos pos) __attribute__ ((pure));

static Pos inv(Pos pos) {
    return POS(pos.y, pos.x);
}

static int is_center(Pos pos) __attribute__ ((pure));

static int is_center(Pos pos) {
    return pos.x == BOARD_SIZE / 2 && pos.y == BOARD_SIZE / 2;
}

typedef struct {
    char word[WORD_SIZE];
    Pos pos, dir;
    int score;
} Entry;

typedef struct {
    char grid[BOARD_SIZE][BOARD_SIZE];
    Entry entries[ENTRY_COUNT];
    char hand[HAND_SIZE];
    int points_used;
    int tiles_used;
    void *parent;
    int score;
    int count;
    int award;
    char *bag;    
} Board;

typedef struct {
    char buffer[WORD_SIZE];
    int index;
} Buffer;

typedef struct {
    Buffer hand;
    Buffer pattern;
    Buffer sequence;
} Buffers;

typedef struct {
    Board *parent;
    Pos pos;
    Pos dir;
} Place;

static int letter_score[] = {
    1, 3, 3,  2, 1, 4, 2, 4, 1, 8, 5, 1, 3,
    1, 1, 3, 10, 1, 1, 1, 1, 4, 4, 8, 4, 10
};

static void try(Board *parent);
static void init_board(Board *board);
static long payout_per_point(Board *board);
static Entry *last_move_entry(Board *board);
static void try_sequence_loop(Place *place, Buffers *ptr);
static long award_desires(Board *board);
static long award_hand(Board *board);

static int is_letter(char c) {
    return (c >= 'A' && c <= 'Z');
}

static int letter_value(char c) {
    return is_letter(c) ? letter_score[c - 'A'] : 0;
}

static int is_lowercase(char c) {
    return (c >= 'a' && c <= 'z');
}

static int word_value(const char *word) {
    int index, value = 0;
    for (index = 0; index < strlen(word); index++) {
	value += letter_value(word[index]);
    }
    return value;
}

static int letter_premium[BOARD_SIZE][BOARD_SIZE] = {
    {0,0,0,1,0,0,0,0,0,0,0,1,0,0,0},
    {0,0,0,0,0,2,0,0,0,2,0,0,0,0,0},
    {0,0,0,0,0,0,1,0,1,0,0,0,0,0,0},
    {1,0,0,0,0,0,0,1,0,0,0,0,0,0,1},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,2,0,0,0,2,0,0,0,2,0,0,0,2,0},
    {0,0,1,0,0,0,1,0,1,0,0,0,1,0,0},
    {0,0,0,1,0,0,0,0,0,0,0,1,0,0,0},
    {0,0,1,0,0,0,1,0,1,0,0,0,1,0,0},
    {0,2,0,0,0,2,0,0,0,2,0,0,0,2,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {1,0,0,0,0,0,0,1,0,0,0,0,0,0,1},
    {0,0,0,0,0,0,1,0,1,0,0,0,0,0,0},
    {0,0,0,0,0,2,0,0,0,2,0,0,0,0,0},
    {0,0,0,1,0,0,0,0,0,0,0,1,0,0,0},
};

static int word_premium[BOARD_SIZE][BOARD_SIZE] = {
    {3,1,1,1,1,1,1,3,1,1,1,1,1,1,3},
    {1,2,1,1,1,1,1,1,1,1,1,1,1,2,1},
    {1,1,2,1,1,1,1,1,1,1,1,1,2,1,1},
    {1,1,1,2,1,1,1,1,1,1,1,2,1,1,1},
    {1,1,1,1,2,1,1,1,1,1,2,1,1,1,1},
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
    {3,1,1,1,1,1,1,2,1,1,1,1,1,1,3},
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
    {1,1,1,1,2,1,1,1,1,1,2,1,1,1,1},
    {1,1,1,2,1,1,1,1,1,1,1,2,1,1,1},
    {1,1,2,1,1,1,1,1,1,1,1,1,2,1,1},
    {1,2,1,1,1,1,1,1,1,1,1,1,1,2,1},
    {3,1,1,1,1,1,1,3,1,1,1,1,1,1,3},
};

static int get_multiplier(int (*table)[BOARD_SIZE], Pos pos) {
    return table[pos.x][pos.y];
}

static int letter_multiplier(Pos pos) {
    return get_multiplier(letter_premium, pos);
}

static int word_multiplier(Pos pos) {
    return get_multiplier(word_premium, pos);
}

static int is_premium_square(Pos pos) {
    return word_multiplier(pos) > 1 || letter_multiplier(pos) > 0;
}

static int is_horizontal(Entry *entry) {
    return entry->dir.x;
}

static void print_entry(Entry *entry) {
    char coordinates[16];
    if (is_horizontal(entry)) {
	sprintf(coordinates, "%d%c", entry->pos.y + 1, entry->pos.x + 'A');
    }
    else {
	sprintf(coordinates, "%c%d", 'A' + entry->pos.x, entry->pos.y + 1);
    }
    printf("%s %s %d,\n", coordinates, entry->word, entry->score);
}

static int contest_score(Board *board) {
    return board->score - BAG_VALUE + board->points_used;
}

static void print_board(Board *board) {
    int i, j;
    printf("    A B C D E F G H I J K L M N O\n");
    printf("  +-------------------------------+\n");
    for (j = 0; j < BOARD_SIZE; j++) {
	printf("%2d|", j + 1);
	for (i = 0; i < BOARD_SIZE; i++) {
	    printf(" %c", board->grid[i][j] ?: '.');
	}
	printf(" |\n");
    }
    printf("  +-------------------------------+\n");
    printf("\n%c:\n", bag[0]);
    for (i = 0; i < board->count; i++) {
	print_entry(board->entries + i);
    }
    printf("\nSCORE: %d", contest_score(board));
    printf("\nHAND: ");
    for (i = 0; i < HAND_SIZE; i++) {
	if (board->hand[i]) printf("%c ", board->hand[i]);
    }
    printf("\nBAG: ");
    for (i = 0; i < BAG_SIZE; i++) {
	if (board->bag[i]) printf("%c ", board->bag[i]); else break;
    }
    printf("\n\n");
    fflush(NULL);
}

static void upcase(char *str) {
    int i = 0;
    while (str[i]) {
	str[i] = toupper(str[i]);
	i++;
    }
}

static void read_lexicon(void) {
    char *status;
    char buffer[MAX_WORD];
    memset((void *) lexicon, 0, LEXICON_SIZE * WORD_SIZE);
    FILE *file = fopen("lexicon.txt", "r");
    do {
	status = fgets(buffer, MAX_WORD, file);
	if (status) {
	    int length = strlen(buffer) - 1;
	    buffer[length] = 0;
	    upcase(buffer);
	    if (length < WORD_SIZE) {
		strcpy(lexicon[lexicon_size++], buffer);
	    }
	}
    } while (status);
    fclose(file);
}

struct Word {
    int count;
    char *word;
    struct Word *children[ALPHABET];
};

typedef struct Word Word;

int words_in_tree = 0;
Word word_tree[WORD_TREE_SIZE];

static Word *alloc_node(void) {
    if (words_in_tree == WORD_TREE_SIZE) {
	printf("ERROR: please increase word tree size");
	exit(-ERANGE);
    }
    return word_tree + (words_in_tree++);
}

static void build_tree(Word *node, int depth, char *word) {
    if (word[depth] == 0) {
	node->word = word;
    }
    else {
	int index = word[depth] - 'A';
	if (!node->children[index]) {
	    node->children[index] = alloc_node();
	}
	if (node->word) node->count++;
	build_tree(node->children[index], depth + 1, word);
    }
}

static void lexicon_tree(void) {
    int index;
    memset(word_tree, 0, sizeof(word_tree));
    alloc_node(); /* reserve root node */
    for (index = 0; index < lexicon_size; index++) {
	build_tree(word_tree, 0, lexicon[index]);
    }
}

static Word *is_word_in_tree(Word *node, int depth, char *word) {
    if (node == NULL) {
	return NULL;
    }
    else if (word[depth] == 0) {
	return node;
    }
    else {
	node = node->children[word[depth] - 'A'];
	return is_word_in_tree(node, depth + 1, word);
    }
}

static Word *is_valid_node(char *word) {
    char tmp[WORD_SIZE];
    upcase(strcpy(tmp, word));
    return is_word_in_tree(word_tree, 0, tmp);
}

static Word *is_valid_word(char *word) {
    Word *node = is_valid_node(word);
    return node && node->word ? node : NULL;
}

static void read_bag(char id) {
    char *status;
    char buffer[256];
    FILE *file = fopen("bags.txt", "r");
    do {
	status = fgets(buffer, sizeof(buffer), file);
	if (status && buffer[0] == id) {
	    buffer[strlen(buffer) - 1] = 0;
	    strcpy(bag, buffer + 3);
	    fclose(file);
	    return;
	}
    } while (status);
    printf("ERROR: bag '%c' not found\n", id);
    exit(-EIO);
}

static int is_on_board(Pos pos) {
    return pos.x >= 0 && pos.x < BOARD_SIZE
	&& pos.y >= 0 && pos.y < BOARD_SIZE;
}

static char get_tile(Board *board, Pos pos) {
    return board->grid[pos.x][pos.y];
}

static void set_tile(Board *board, Pos pos, char letter) {
    board->grid[pos.x][pos.y] = letter;
}

static int is_occupied(Board *board, Pos pos) {
    return is_on_board(pos) && get_tile(board, pos);
}

static int is_desolate(Board *board, Pos pos) {
    return !is_center(pos)
	&& !is_occupied(board, pos)
	&& !is_occupied(board, sub(pos, POS(1, 0)))
	&& !is_occupied(board, add(pos, POS(1, 0)))
	&& !is_occupied(board, sub(pos, POS(0, 1)))
	&& !is_occupied(board, add(pos, POS(0, 1)));
}

static Pos seek_start(Board *board, Pos pos, Pos dir) {
    while (is_occupied(board, sub(pos, dir))) {
	pos = sub(pos, dir);
    }
    return pos;
}

static int is_occupied_near(Board *board, Pos pos, Pos dir) {
    return is_occupied(board, sub(pos, dir))
	|| is_occupied(board, add(pos, dir));
}

static Pos get_word_at_pos(Board *board, Pos pos, Pos dir, char *word) {
    pos = seek_start(board, pos, dir);

    int index = 0;
    Pos head = pos;
    while (is_occupied(board, pos)) {
	word[index] = get_tile(board, pos);
	pos = add(pos, dir);
	index++;
    }

    word[index] = 0;
    return head;
}

static int validate_pos_in_direction(Board *board, Pos pos, Pos dir) {
    char word[WORD_SIZE];
    get_word_at_pos(board, pos, dir, word);
    return is_valid_word(word) ? word_value(word) : 0;
}

static void add_entry(Board *board, Pos pos, Pos dir, int score) {
    if (board->count == ENTRY_COUNT) {
	printf("ERROR: reached ENTRY_COUNT, please increase\n");
	exit(-ERANGE);
    }
    else {
	Entry *entry = board->entries + board->count;
	entry->pos = get_word_at_pos(board, pos, dir, entry->word);
	entry->score = score;
	entry->dir = dir;
	board->count++;
    }
}

static Board desires;
static char desire_resources[ALPHABET];
static int desire_map[BOARD_SIZE][BOARD_SIZE];

static char *channels[BOARD_SIZE] = {
    "O . V . . . . . . . . . . . E",
    "X . O . . . . . . . . . W . M",
    "Y A w L . . . . . . . . A G O",
    "P I . I . . . . . . . . W E T",
    "H M . . . . . . . . . . . T I",
    "E . . . . V . . . . . . . . O",
    "N I T . L I . F . . . . B A N",
    "B . O P I A T E . . . H A L L",
    "U D O . . . . R A I D E D . E",
    "T I . . R U I N . . . . E F S",
    "A G R E E D . Y . . . . . o S",
    "Z . E R G O . . . . . Q . U N",
    "O R C A . . . . . . N A C R E",
    "N . K . . . . . . J U T . . S",
    "E . . . . . . . . . . . . . S",
};

static char get_channel(Pos pos) {
    return channels[pos.y][pos.x * 2];
}

static int is_forbiden_move(Board *board) {
    int x, y, mismatch = 0;
    for (y = 0; y < BOARD_SIZE; y++) {
	for (x = 0; x < BOARD_SIZE; x++) {
	    Pos pos = POS(x, y);
	    char need = toupper(get_channel(pos));
	    char have = toupper(get_tile(board, pos));
	    if (have && need != have) mismatch++;
	}
    }
    return mismatch > 0;
}

static void add_desire(char *word, Pos pos, Pos dir) {
    int index = 0;
    Pos start = pos;
    while (word[index]) {
	char letter = get_tile(&desires, pos);
	if (letter && letter != word[index]) {
	    printf("ERROR: conflicting desires\n");
	    exit(-EINVAL);
	}
	else {
	    set_tile(&desires, pos, word[index]);
	}
	pos = add(pos, dir);
	index++;
    }
    add_entry(&desires, start, dir, 0);
}

static int get_desire_value(Pos pos) {
    return is_on_board(pos) ? desire_map[pos.x][pos.y] : 0;
}

static int max_desire_value_around(Pos pos) {
    int max = 0;
    max = MAX(max, get_desire_value(add(pos, POS(1, 0))));
    max = MAX(max, get_desire_value(sub(pos, POS(1, 0))));
    max = MAX(max, get_desire_value(add(pos, POS(0, 1))));
    max = MAX(max, get_desire_value(sub(pos, POS(0, 1))));
    return max;
}

static void iterate_over_desire_map(void) {
    int x, y;
    int buf[BOARD_SIZE][BOARD_SIZE];
    for (y = 0; y < BOARD_SIZE; y++) {
	for (x = 0; x < BOARD_SIZE; x++) {
	    int now = desire_map[x][y];
	    int max = max_desire_value_around(POS(x, y));
	    buf[x][y] = (now == 0 && max > 0) ? max + 1 : now;
	}
    }
    for (y = 0; y < BOARD_SIZE; y++) {
	for (x = 0; x < BOARD_SIZE; x++) {
	    desire_map[x][y] = buf[x][y];
	}
    }
}

static void mark_desire(int desire_num) {
    int index = 0;
    Entry *entry = desires.entries + desire_num;
    Pos pos = entry->pos;
    while (entry->word[index]) {
	desire_map[pos.x][pos.y] = 1;
	pos = add(pos, entry->dir);
	index++;
    }
}

static void populate_desire_map(void) {
    int index;
    for (index = 0; index < desires.count; index++) {
	mark_desire(index);
    }
    for (index = 0; index < DOUBLE_SIZE; index++) {
	iterate_over_desire_map();
    }
}

static void calculate_desire_resources(void) {
    int x, y;
    memset(desire_resources, 0, ALPHABET);
    memset(desire_map, 0, sizeof(desire_map));
    for (x = 0; x < BOARD_SIZE; x++) {
	for (y = 0; y < BOARD_SIZE; y++) {
	    char letter = get_tile(&desires, POS(x, y));
	    if (is_letter(letter)) desire_resources[letter - 'A']++;
	}
    }
    populate_desire_map();
}

static int validate_accross(Board *board, Pos pos, Pos dir) {
    dir = inv(dir);
    char word[WORD_SIZE];
    get_word_at_pos(board, pos, dir, word);
    return strlen(word) == 1 || is_valid_word(word);
}

static int check_if_desire_is_profitable(Board *board, Entry *desire) {
    int index = 0;
    int used_premiums = 0;
    int total_premiums = 0;
    Pos pos = desire->pos;
    while (desire->word[index]) {
	if (is_premium_square(pos)) {
	    total_premiums++;
	    if (get_tile(board, pos)) {
		used_premiums++;
	    }
	}
	pos = add(pos, desire->dir);
	index++;
    }
    return used_premiums == 0 || used_premiums == total_premiums;
}

static int validate_desire_in_direction(Board *board, Entry *desire) {
    return validate_pos_in_direction(board, desire->pos, desire->dir);
}

static int significant_value(char letter) {
    return letter >= 'a' && letter <= 'z'
	&& letter_value(toupper(letter)) > allow_blanks;
}

static int can_fulfill(Board *board, Entry *desire, char *resources) {
    int index = 0;
    Pos pos = desire->pos;
    Board scratch = *board;
    char *word = desire->word;
    while (word[index]) {
	char letter = get_tile(&scratch, pos);
	if (letter == 0) {
	    set_tile(&scratch, pos, word[index]);
	}
	else if (toupper(letter) != word[index] || significant_value(letter)) {
	    return 0;
	}
	else {
	    resources[word[index] - 'A']--;
	}
	if (!validate_accross(&scratch, pos, desire->dir)) {
	    return 0;
	}
	pos = add(pos, desire->dir);
	index++;
    }
    return validate_desire_in_direction(&scratch, desire)
	&& check_if_desire_is_profitable(board, desire);
}

static void update_available(char *buf, int size, char *available) {
    int index;
    for (index = 0; index < size; index++) {
	if (is_letter(buf[index])) available[buf[index] - 'A']++;
    }
}

static int fulfillment_possible(Board *board) {
    int index;
    char resources[ALPHABET];
    char available[ALPHABET];
    memset(available, 0, ALPHABET);
    memcpy(resources, desire_resources, ALPHABET);
    for (index = 0; index < desires.count; index++) {
	if (!can_fulfill(board, desires.entries + index, resources)) {
	    return 0;
	}
    }
    update_available(board->bag, strlen(board->bag), available);
    update_available(board->hand, HAND_SIZE, available);
    for (index = 0; index < ALPHABET; index++) {
	if (available[index] < resources[index]) {
	    return 0;
	}
    }
    return 1;
}

static int desire_tile_score(Pos pos) {
    int distance = get_desire_value(pos);
    if (distance <= 6) {
	return (1 << (28 - 4 * distance));
    }
    else if (distance <= 20) {
	return 22 - distance;
    }
    else {
	return 1;
    }
}

static long award_desires(Board *board) {
    return board->award;
}

static void fill_hand(Board *board, int *hand) {
    int index;
    memset(hand, 0, ALPHABET * sizeof(int));
    for (index = 0; index < HAND_SIZE; index++) {
	char letter = board->hand[index];
	if (letter && letter != '?') hand[letter - 'A']++;
    }
}

static int count_blanks(Board *board) {
    int index, count = 0;
    for (index = 0; index < HAND_SIZE; index++) {
	if (board->hand[index] == '?') count++;
    }
    return allow_blanks ? count : 0;
}

static int desire_completion(Board *board, Entry *entry) {
    int used = 0;
    int index = 0;
    int in_hand = 0;
    int hand[ALPHABET];
    Pos pos = entry->pos;
    fill_hand(board, hand);
    while (entry->word[index]) {
	if (get_tile(board, pos)) {
	    used++;
	}
	else {
	    char letter = entry->word[index];
	    if (hand[letter - 'A'] > 0) {
		hand[letter - 'A']--;
		in_hand++;
	    }
	}
	pos = add(pos, entry->dir);
	index++;
    }
    return used == index ? 0 : in_hand + used + count_blanks(board);
}

static long award_hand(Board *board) {
    int index, max_completion = 0;
    for (index = 0; index < desires.count; index++) {
	Entry *entry = desires.entries + index;
	int current_completion = desire_completion(board, entry);
	max_completion = MAX(max_completion, current_completion);
    }
    return max_completion;
}

static int manhattan(Pos a, Pos b) {
    return abs(a.x - b.x) + abs(a.y - b.y);
}

static int inverse_distance_of_pos(Board *board, Pos pos) {
    int x, y, distance = DOUBLE_SIZE;
    for (y = 0; y < BOARD_SIZE; y++) {
	for (x = 0; x < BOARD_SIZE; x++) {
	    Pos here = POS(x, y);
	    if (get_tile(board, here)) {
		distance = MIN(distance, manhattan(pos, here));
	    }
	}
    }
    return DOUBLE_SIZE - distance;
}

static int distance_of_desire(Board *board, Entry *entry) {
    int index = 0;
    int score = 0;
    Pos pos = entry->pos;
    while (entry->word[index]) {
	score += inverse_distance_of_pos(board, pos);
	pos = add(pos, entry->dir);
	index++;
    }
    return score;
}

static long desire_distances(Board *board) {
    int index;
    long score = 0;
    for (index = 0; index < board->count; index++) {
	score += distance_of_desire(board, desires.entries + index);
    }
    return score;
}

static int get_letter_bonus(Board *board, Pos pos) {
    return letter_value(get_tile(board, pos)) * letter_multiplier(pos);
}

typedef long (*Goodness)(Board *board);

static long score_on_board(Board *board) {
    return (long) board->score;
}

static long safe_div(long dividend, long divisor) {
    return divisor == 0 ? 0 : dividend / divisor;
}

static long payout_per_point(Board *board) {
    return safe_div(1000 * board->score, board->points_used);
}

static Entry *last_move_entry(Board *board) {
    static Entry bogus_entry = { .score = 0, .word = "" };
    return board->count ? &board->entries[board->count - 1] : &bogus_entry;
}

static Goodness metrics[] = {
    score_on_board,
    payout_per_point,
    desire_distances,
    award_desires,
    award_hand,
};

typedef struct {
    long score;
    Board board;
} Cutoff;

#define METRIC_COUNT ARRAY_SIZE(metrics)
static Cutoff cutoff_table[BAG_SIZE][METRIC_COUNT][METRIC_COUNT][POOL_SIZE];

static int sign(long value) {
    return (value > 0) ? 1 : ((value < 0) ? -1 : 0);
}

static int is_better(Goodness metric, Board *board, Cutoff *pool) {
    int index, worst = 0;
    long score = metric(board);
    for (index = 1; index < POOL_SIZE; index++) {
	if (pool[index].score < pool[worst].score) worst = index;
    }
    int better = sign(score - pool[worst].score);
    if (better > 0) {
	pool[worst].score = score;
	pool[worst].board = *board;
    }
    return better;
}

static int is_interesting(Board *board) {
    int index, secondary;
    int depth = board->tiles_used - 1;
    for (index = 0; index < METRIC_COUNT; index++) {
	Cutoff *pool = cutoff_table[depth][index][index];
	int better = is_better(metrics[index], board, pool);
	for (secondary = 0; secondary < ARRAY_SIZE(metrics); secondary++) {
	    pool = cutoff_table[depth][index][secondary];
	    if (index == secondary) {
		continue;
	    }
	    else if (better > 0) {
		memset(pool, 0, sizeof(Cutoff) * POOL_SIZE);
	    }
	    else if (better == 0) {
		if (is_better(metrics[secondary], board, pool) > 0) return 1;
	    }
	}
	if (better > 0) return 1;
    }
    return 0;
}

static int same_boards(Board *board1, Board *board2) {
    return memcmp(board1->grid, board2->grid, BOARD_SIZE * BOARD_SIZE) == 0;
}

static int is_clone(Board *board) {
    int m1, m2, pool;
    int depth = board->tiles_used - 1;
    for (m1 = 0; m1 < METRIC_COUNT; m1++) {
	for (m2 = 0; m2 < METRIC_COUNT; m2++) {
	    for (pool = 0; pool < POOL_SIZE; pool++) {
		Board *candidate = &cutoff_table[depth][m1][m2][pool].board;
		if (same_boards(board, candidate)) return 1;
	    }
	}
    }
    return 0;
}

static Board champion;
static int restrain = 0;

static void evaluate(Board *board) {
    if (is_clone(board)) return;

    if (!fulfillment_possible(board)) return;

    if (restrain && is_forbiden_move(board)) return;

    if (contest_score(board) > contest_score(&champion)) {
	champion = *board;
	print_board(board);
    }
    if (restrain || is_interesting(board)) {
	try(board);
    }
}

static void replenish(Board *board) {
    int index;
    for (index = 0; index < HAND_SIZE; index++) {
	if (*board->bag == 0) {
	    return;
	}
	else if (board->hand[index] == 0) {
	    board->hand[index] = *board->bag;
	    board->bag++;
	}
    }
}

static void update_stats(Board *board, Board *parent) {
    board->score += last_move_entry(board)->score;
    board->parent = parent;
}

static void discard_letter(Board *board, char letter) {
    int index;
    for (index = 0; index < HAND_SIZE; index++) {
	char tile = board->hand[index];
	if (tile == letter || (tile == '?' && is_lowercase(letter))) {
	    board->points_used += letter_value(letter);
	    board->hand[index] = 0;
	    board->tiles_used++;
	    return;
	}
    }
    printf("ERROR: cound not discard '%c'\n", letter);
    exit(-EINVAL);
}

static void try_sequence(Board *parent, Pos pos, Pos dir, char *sequence) {
    int length = strlen(sequence);
    Board copy = *parent;
    Board *board = &copy;
    int index, score = 0;
    int is_valid_move = 0;
    int auxilary_score = 0;
    int total_multiplier = 1;

    is_valid_move = is_occupied(board, sub(pos, dir));

    for (index = 0; index < length; index++) {
	if (!is_on_board(pos)) {
	    return;
	}
	if (is_center(pos)) {
	    is_valid_move = 1;
	}
	if (get_tile(board, pos)) {
	    is_valid_move = 1;
	    index--;
	}
	else {
	    char letter = sequence[index];
	    set_tile(board, pos, letter);
	    int multiplier =  word_multiplier(pos);
	    int tile_bonus = get_letter_bonus(board, pos);
	    board->award += desire_tile_score(pos);
	    total_multiplier *= multiplier;
	    discard_letter(board, letter);
	    score += tile_bonus;

	    Pos inv_dir = inv(dir);
	    if (is_occupied_near(board, pos, inv_dir)) {
		int sub_score = validate_pos_in_direction(board, pos, inv_dir);
		if (sub_score) {
		    auxilary_score += multiplier * (tile_bonus + sub_score);
		    is_valid_move = 1;
		}
		else {
		    return;
		}
	    }
	}
	pos = add(pos, dir);
    }

    is_valid_move = is_valid_move || is_occupied(board, pos);
    int main_word_score = validate_pos_in_direction(board, pos, dir);

    if (is_valid_move && main_word_score) {
	score = auxilary_score + total_multiplier * (score + main_word_score);
	if (length == HAND_SIZE) score += BONUS_SCORE;
	add_entry(board, pos, dir, score);
	update_stats(board, parent);
	replenish(board);
	evaluate(board);
    }
}

static void push(Buffer *ptr, char letter) {
    ptr->buffer[ptr->index] = letter;
    ptr->buffer[ptr->index + 1] = 0;
    ptr->index++;
}

static char peek(Buffer *ptr, int offset) {
    return ptr->buffer[ptr->index + offset];
}

static void poke(Buffer *ptr, int offset, char letter) {
    ptr->buffer[ptr->index + offset] = letter;
}

static int is_valid_pattern_prefix(char *prefix, Word *node) {
    if (*prefix == '.' || *prefix == '*' || *prefix == 0) {
	return 1;
    }
    else {
	node = node->children[toupper(*prefix) - 'A'];
	return !node ? 0 : is_valid_pattern_prefix(prefix + 1, node);
    }
}

static void try_sequence_char(Place *place, Buffers *ptr, char letter) {
    char pattern = peek(&ptr->pattern, -1);
    poke(&ptr->pattern, -1, letter);
    if (is_valid_pattern_prefix(ptr->pattern.buffer, word_tree)) {
	push(&ptr->sequence, letter);
	if (pattern != '.') {
	    Board *parent = place->parent;
	    try_sequence(parent, place->pos, place->dir, ptr->sequence.buffer);
	}
	try_sequence_loop(place, ptr);
	ptr->sequence.index--;
    }
    poke(&ptr->pattern, -1, pattern);
}

static void try_sequence_loop(Place *place, Buffers *ptr) {
    char pattern = peek(&ptr->pattern, 0);
    char *hand = ptr->hand.buffer;
    ptr->pattern.index++;
    if (pattern) {
	if (pattern != '.' && pattern != '*') {
	    try_sequence_loop(place, ptr);
	}
	else {
	    int index;
	    for (index = 0; index < HAND_SIZE; index++) {
		char letter = hand[index];
		hand[index] = 0;
		if (letter == '?') {
		    char substitue;
		    for (substitue = 'a'; substitue <= 'z'; substitue++) {
			try_sequence_char(place, ptr, substitue);
		    }
		}
		else if (letter) {
		    try_sequence_char(place, ptr, letter);
		}
		hand[index] = letter;
	    }
	}
    }
    ptr->pattern.index--;
}

static int get_pattern(Board *parent, Pos pos, Pos dir, char *pattern) {
    int index = 0;
    int empty = 0;
    int all_is_desolate = 1;
    pos = seek_start(parent, pos, dir);
    while (empty < HAND_SIZE && is_on_board(pos)) {
	char letter = get_tile(parent, pos);
	if (letter) {
	    pattern[index] = letter;
	    all_is_desolate = 0;
	}
	else {
	    all_is_desolate = all_is_desolate && is_desolate(parent, pos);
	    pattern[index] = all_is_desolate ? '.' : '*';
	    empty++;
	}
	pos = add(pos, dir);
	index++;
    }
    pattern[index] = 0;
    return !all_is_desolate;
}

static void try_all_sequences(Board *parent, Pos pos, Pos dir) {
    Buffers ptr;
    memset(&ptr, 0, sizeof(ptr));
    if (get_pattern(parent, pos, dir, ptr.pattern.buffer)) {
	memcpy(ptr.hand.buffer, parent->hand, HAND_SIZE);
	Place place = { .parent = parent, .pos = pos, .dir = dir };
	try_sequence_loop(&place, &ptr);
    }
}

static void try(Board *parent) {
    int x, y;
    for (x = BOARD_SIZE - 1; x >= 0; x--) {
	for (y = BOARD_SIZE - 1; y >= 0; y--) {
	    Pos pos = POS(x, y);
	    if (!is_occupied(parent, pos)) {
		try_all_sequences(parent, pos, POS(1, 0));
		try_all_sequences(parent, pos, POS(0, 1));
	    }
	}
    }
}

static void init_board(Board *board) {
    memset(board, 0, sizeof(Board));
    board->bag = bag;
    replenish(board);
}

static void run(void) {
    Board empty;
    init_board(&empty);
    memset(cutoff_table, 0, sizeof(cutoff_table));
    try(&empty);
}

static void desire_megaword(char *word, Pos pos, char *aux_word, Pos aux_pos) {
    init_board(&desires);
    add_desire(word, pos, POS(0, 1));
    if (aux_word) add_desire(aux_word, aux_pos, POS(0, 1));
    calculate_desire_resources();
    run();
}

static int is_legal_subword(const char *word, int offset, int length) {
    int index = 0;
    char subword[WORD_SIZE];
    subword[length] = 0;
    while (index < length) {
	subword[index] = word[offset + index];
	index++;
    }
    return is_valid_word(subword) != NULL ? 1 : 0;
}

static int good_fragments(const char *word) {
    return is_legal_subword(word, 1, 2)
	 + is_legal_subword(word, 4, 3)
	 + is_legal_subword(word, 8, 3)
	 + is_legal_subword(word, 1, 2);
}

static int count_letters(const char *word, char letter) {
    int count = 0, index = 0;
    while (word[index]) {
	if (word[index] == letter) count++;
	index++;
    }
    return count;
}

static long letters_remaining(char c) {
    return count_letters(bag, c) - count_letters("OXYPHENBUTAZONE", c);
}

static long bag_ratio(const char *word) {
    char letter;
    long ratio = 0;
    for (letter = 'A'; letter <= 'Z'; letter++) {
	int count = count_letters(word, letter);
	if (count) ratio += letters_remaining(letter) / count;
    }
    return ratio;
}

static int compare_words(const char **a, const char **b) {
    return good_fragments(*a) < good_fragments(*b);
    return bag_ratio(*a) < bag_ratio(*b);
}

static void desire_two_megawords(char *word1, char *word2) {
    desire_megaword(word1, POS(0, 0), word2, POS(14, 0));
    desire_megaword(word1, POS(14, 0), word2, POS(0, 0));
}

int main(int argc, char **argv) {
    read_bag(argc > 1 && argv[1][0] != 0 ? argv[1][0] : 'A');
    read_lexicon();
    lexicon_tree();

    init_board(&champion);
    char *megawords[4096];
    int index, count = 0;
    for (index = 0; index < lexicon_size; index++) {
	char *word = lexicon[index];
	if (strlen(word) == BOARD_SIZE) {
	    megawords[count] = word;
	    count++;
	}
    }

    qsort(megawords, count, sizeof(char *), (void *) &compare_words);

    for (index = 0; index < count; index++) {
	desire_two_megawords("OXYPHENBUTAZONE", megawords[index]);
    }
    return 0;
}
